/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

MF_COND(WITH_DSPAM)

#include "srvcfg.h"
#undef HAVE_CONFIG_H
#include <libdspam.h>
#include "mflib/dspam.h"
#include "msg.h"

MF_VAR(dspam_home, STRING, SYM_PRECIOUS);
MF_VAR(dspam_user, STRING, SYM_PRECIOUS);
MF_VAR(dspam_group, STRING, SYM_PRECIOUS);
#define DEFAULT_DSPAM_HOME_STR "dspam"
#define DEFAULT_DSPAM_HOME_LEN (sizeof(DEFAULT_DSPAM_HOME_STR)-1)
MF_VAR(dspam_probability, NUMBER);
MF_VAR(dspam_confidence, NUMBER);
MF_VAR(dspam_prec, NUMBER);
#define DEFAULT_DSPAM_PREC 3

static int _dspam_initialized;

static void
_dspam_shutdown()
{
	dspam_shutdown_driver(NULL);
}

struct transtab
{
	int trans_from;
	int trans_to;
};

static struct transtab mode_trans[] = {
	{ _MF_DSM_PROCESS, DSM_PROCESS },
	{ _MF_DSM_CLASSIFY, DSM_CLASSIFY }
};

static struct transtab flag_trans[] = {
	{ _MF_DSF_SIGNATURE,  DSF_SIGNATURE },
	{ _MF_DSF_NOISE,      DSF_NOISE },    
	{ _MF_DSF_WHITELIST,  DSF_WHITELIST }
};

static struct transtab tokenizer_trans[] = {
	{ _MF_DSZ_WORD,  DSZ_WORD },
	{ _MF_DSZ_CHAIN, DSZ_CHAIN },
	{ _MF_DSZ_SBPH,  DSZ_SBPH },
	{ _MF_DSZ_OSB,   DSZ_OSB }
};

static struct transtab tmod_trans[] = {
	{ _MF_DST_TEFT, DST_TEFT },
	{ _MF_DST_TOE,  DST_TOE },
	{ _MF_DST_TUM,  DST_TUM }
};

static struct transtab class_trans[] = {
	{ _MF_DSR_ISSPAM,     DSR_ISSPAM },    
	{ _MF_DSR_ISINNOCENT, DSR_ISINNOCENT },
	{ _MF_DSR_NONE,       DSR_NONE }
};

static struct transtab source_trans[] = {
	{ _MF_DSS_ERROR,       DSS_ERROR },     
	{ _MF_DSS_CORPUS,      DSS_CORPUS },    
	{ _MF_DSS_INOCULATION, DSS_INOCULATION },
	{ _MF_DSS_NONE,        DSS_NONE }      
};

static int
translate(struct transtab *tab, size_t count, int num, int *ret)
{
	for (; count; tab++, count--)
		if (tab->trans_from == num) {
			*ret = tab->trans_to;
			return 0;
		}
	return -1;
}

static int
translate_flags(struct transtab *tab, size_t count, int num)
{
	int f = 0;
	for (; count; tab++, count--)
		if (tab->trans_from & num)
			f |= tab->trans_to;
	return f;
}

static void
stream_cleanup(void *ptr)
{
	mu_stream_t str = ptr;
	mu_stream_unref(str);
}

static void
ctx_cleanup(void *ptr)
{	
	DSPAM_CTX *ctx = ptr;
	dspam_destroy(ctx);
}

/* number dspam(number msg, number flags; number class_source) */
MF_DSEXP
MF_DEFUN(dspam, NUMBER, NUMBER nmsg, NUMBER mode_flags, OPTIONAL, NUMBER class_src)
{
	int rc, n;
	DSPAM_CTX *ctx;               	/* DSPAM Context */
	int mode;
	int flags;
	mu_message_t msg;
	mu_stream_t msgstr, instr;
	const char *msgbuf;
	size_t msgsize;
	unsigned prec;
	mu_transport_t trans[2];
	
	/* Prepare message buffer */
	msg = bi_message_from_descr(env, nmsg);
	rc = mu_message_size(msg, &msgsize);
	MF_ASSERT(rc == 0,
		  mfe_failure,
 		  "mu_message_size: %s", mu_strerror(rc));
	
	rc = mu_memory_stream_create(&msgstr, MU_STREAM_RDWR);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		 "mu_static_memory_stream_create: %s",
		  mu_strerror(rc));
	MF_DCL_CLEANUP(msgstr, stream_cleanup);

	rc = mu_message_get_streamref(msg, &instr);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_message_get_streamref: %s",
		  mu_strerror(rc));
	MF_DCL_CLEANUP(instr, stream_cleanup);

	rc = mu_stream_copy(msgstr, instr, msgsize, NULL);
	MF_ASSERT(rc == 0,
		  mfe_failure,
		  "mu_stream_copy: %s",
		  mu_strerror(rc));
	
	MF_CLEANUP(instr);

	mu_stream_ioctl(msgstr, MU_IOCTL_TRANSPORT, MU_IOCTL_OP_GET, trans);
	msgbuf = (const char*)trans[0];
	
	/* Initialize dspam library and set up global variables, if
	   needed */
	if (!_dspam_initialized) {
		size_t off, len;
		char *s;
		
		dspam_init_driver(NULL);
		atexit(_dspam_shutdown);
		_dspam_initialized = 1;

		if (MF_VAR_REF(dspam_home) == NULL) {
			len = strlen(mailfromd_state_dir);
			while (len > 0 && mailfromd_state_dir[len-1] == '/')
				len--;
			s = MF_ALLOC_HEAP(off,
					  len + DEFAULT_DSPAM_HOME_LEN + 2);
			memcpy(s, mailfromd_state_dir, len);
			s[len++] = '/';
			strcpy(s + len, DEFAULT_DSPAM_HOME_STR);
			MF_VAR_REF(dspam_home, off);
		}

		if (MF_VAR_REF(dspam_user) == NULL)
			MF_VAR_SET_STRING(dspam_user, mf_server_user);

		if (MF_VAR_REF(dspam_prec) == 0)
			MF_VAR_REF(dspam_prec, DEFAULT_DSPAM_PREC);
	}

	/* Prepare DSPAM context */
	MF_ASSERT(translate(mode_trans, MU_ARRAY_SIZE(mode_trans),
			    mode_flags & _MF__DSM_MASK, &mode) == 0,
		  mfe_failure,
		  "bad dspam mode");
	flags = translate_flags(flag_trans, MU_ARRAY_SIZE(flag_trans),
				mode_flags);
	
	/* Initialize the DSPAM context */
	ctx = dspam_init(MF_VAR_STRING(dspam_user),
			 MF_VAR_STRING(dspam_group),
			 MF_VAR_STRING(dspam_home), mode,
			 flags);
	MF_ASSERT(ctx != NULL,
		  mfe_failure,
		  "dspam_init failed");
	MF_DCL_CLEANUP(ctx, ctx_cleanup);
	
	/* Use graham and robinson algorithms, graham's p-values */
	/* FIXME: Get from args? */
	ctx->algorithms = DSA_GRAHAM | DSA_BURTON | DSP_GRAHAM;

	/* Configure tokenizer */
	MF_ASSERT(translate(tokenizer_trans, MU_ARRAY_SIZE(tokenizer_trans),
			    mode_flags & _MF__DSZ_MASK, &ctx->tokenizer) == 0,
		  mfe_failure,
		  "bad dspam tokenizer");
	/* Set training mode */
	MF_ASSERT(translate(tmod_trans, MU_ARRAY_SIZE(tmod_trans),
			    mode_flags & _MF__DST_MASK, &ctx->training_mode) == 0,
		  mfe_failure,
		  "bad dspam tokenizer");
	
	/* Set up classification and source */
	if (n = MF_OPTVAL(class_src)) {
		MF_ASSERT(translate(class_trans, MU_ARRAY_SIZE(class_trans),
				    n & _MF__DSR_MASK,
				    &ctx->classification) == 0,
			  mfe_failure,
			  "bad dspam classification flag");
		MF_ASSERT(translate(source_trans, MU_ARRAY_SIZE(source_trans),
				    n & _MF__DSS_MASK,
				    &ctx->source) == 0,
			  mfe_failure,
			  "bad dspam source flag");
	}
	
	/* Process the message */
	MF_ASSERT(dspam_process(ctx, msgbuf) == 0,
		  mfe_failure,
		  "dspam_process failed");

	rc = (unsigned) MF_VAR_REF(dspam_prec);
	prec = 1;
	while (rc--)
		prec *= 10;
	MF_VAR_REF(dspam_probability,
		   (unsigned long) (ctx->probability * prec));
	MF_VAR_REF(dspam_confidence,
		   (unsigned long) (ctx->confidence * prec));
	rc = ctx->result;
	MF_CLEANUP(ctx);
	
	/* FIXME: Any additional processing? */

	MF_RETURN(rc);
}
END

MF_INIT
